
#include "shared.h"

/* System memory */
uint32 _arm_aligned_2;
uint8 ram[0x8000];      /* Work RAM */
uint8 cdram[0x10000];   /* CD unit RAM (64k) */
uint8 bram[0x2000];     /* Backup RAM (8K) */
uint8 rom[0x100000];    /* HuCard ROM (1MB) */
uint8 save_bram;        /* 1= BRAM registers were accessed */
#ifdef FAST_MEM
uint32 _arm_aligned_3;
uint8 dummy[0x2000];    /* Dummy block for unknown access */
uint8 *read_ptr[8];     /* Memory read pointers */
uint8 *write_ptr[8];    /* Memory write pointers */
#endif

/* I/O port data */
uint8 joy_sel = 0;
uint8 joy_clr = 0;
uint8 joy_cnt = 0;


/*--------------------------------------------------------------------------*/
/* Init, reset, shutdown functions                                          */
/*--------------------------------------------------------------------------*/

int pce_init(void)
{
#ifdef FAST_MEM
    bank_reset();
#endif
    h6280_reset(0);
    h6280_set_irq_callback(&pce_irq_callback);
    return (1);
}

void pce_reset(void)
{
    joy_sel = joy_clr = joy_cnt = 0;
    memset(ram, 0, 0x8000);
    memset(cdram, 0, 0x10000);
#ifdef FAST_MEM
    memset(dummy, 0, 0x2000);
    bank_reset();
#endif
//    load_file("pce.brm", (char*)bram, 0x2000);
    h6280_reset(0);
    h6280_set_irq_callback(&pce_irq_callback);
}

void pce_shutdown(void)
{
	char tmp_str[255];
	GetStateFileName(-1,tmp_str);
    if(save_bram) save_file(tmp_str, (char*)bram, 0x2000);
#ifdef DEBUG
    error("PC:%04X\n", h6280_get_pc());
#endif    
}

/*--------------------------------------------------------------------------*/
/* Memory handlers                                                          */
/*--------------------------------------------------------------------------*/

#ifndef FAST_MEM

void cpu_writeport16(int port, int data)
{
    vdc_w(port, data);
}

void cpu_writemem21(int address, int data)
{
    uint8 page = (address >> 13) & 0xFF;

    /* RAM (F8) */
    if(page == 0xF8 || page == 0xF9 || page == 0xFA || page == 0xFB) {
        ram[(address & 0x7FFF)] = data;
        return;
    }

    /* I/O (FF) */
    if(page == 0xFF) {
        io_page_w(address & 0x1FFF, data);
        return;
    }

    /* CD RAM */
    if((page >= 0x80) && (page <= 0x87)) {
        cdram[(address & 0xFFFF)] = data;
        return;
    }

    /* Backup RAM (F7) */
    if(page == 0xF7) {
        bram[(address & 0x1FFF)] = data;
        return;
    }

#ifdef DEBUG
    error("write %02X to %02X:%04X (%08X)\n", data, page, address & 0x1fff, h6280_get_reg(H6280_PC));
#endif
}

int cpu_readmem21(int address)
{
    uint8 page;

    if(address <= 0x0FFFFF) return (rom[(address)]);

    page = (address >> 13) & 0xFF;

    /* ROM (00-7F) */
    if(page <= 0x7F) return (rom[(address)]);

    /* RAM (F8) */
    if(page == 0xF8 || page == 0xF9 || page == 0xFA || page == 0xFB) return (ram[(address & 0x7FFF)]);

    /* I/O (FF) */
    if(page == 0xFF) return (io_page_r(address & 0x1FFF));

    /* CD RAM */
    if((page >= 0x80) && (page <= 0x87)) return (cdram[(address & 0xFFFF)]);

    /* Backup RAM (F7) */
    if(page == 0xF7) return (bram[(address & 0x1FFF)]);

#ifdef DEBUG
    error("read %02X:%04X (%08X)\n", page, address & 0x1fff, h6280_get_reg(H6280_PC));
#endif
    return (0xFF);
}

#endif

int pce_irq_callback(int irqline)
{
    return (0);
}

/*--------------------------------------------------------------------------*/
/* Hardware page handlers                                                   */
/*--------------------------------------------------------------------------*/
/*****************************/
 /*  yoyofr : from hugo       */
 /*****************************/   
int io_buffer;

void io_page_w(int address, int data)
{
    if ((address >= 0x800) && (address < 0x1800)) // We keep the io buffer value
    	io_buffer = data;
	
    switch(address & 0x1F00)
    {
        case 0x0000: /* VDC */
            address&=3;
            if(address <= 0x0003) { vdc_w(address, data); return; }
            break;

        case 0x0400: /* VCE */
            address&=0x0407;
            if(address <= 0x0405) { vce_w(address, data); return; }
            break;

        case 0x0800: /* PSG */
            address&=0x080F;
            if(address <= 0x0809) { psg_w(address, data); return; };
            break;

        case 0x0C00: /* Timer */
            address&=0xC01;
            if(address == 0x0C00 || address == 0x0C01) { H6280_timer_w(address & 1, data); return; };
            break;

        case 0x1000: /* I/O */
            address&=0x1000;
            if(address == 0x1000) { input_w(data); return; }
            break;

        case 0x1400: /* IRQ control */
            address&=0x140F;
            if(address == 0x1402 || address == 0x1403) { H6280_irq_status_w(address & 1, data); return; };
            break;

        case 0x1800: /* CD-ROM */
#ifdef DEBUG
            error("cdrom %04X = %02X\n", address, data);
#endif
			switch (address & 15)
			{			
            	case 7: save_bram = 1;
/*            	case 8:
				  psg.adpcm_ptr = (psg.adpcm_ptr&0xFF00)|data;
				  return;
				case 9:
				  psg.adpcm_ptr = (psg.adpcm_ptr&0x00FF)|((Int32)(data)<<8);
				  return;
				case 0x0A:
				  PCM[psg.adpcm_wptr++] = data;
				  return;
			  case 0x0D:
				  if ((data & 0x03) == 0x03)
				    {
				      psg.adpcm_dmaptr = psg.adpcm_ptr;	// set DMA pointer
				    }

				  if (data & 0x04)
				    {
				      psg.adpcm_wptr = psg.adpcm_ptr;	// set write pointer		
	    			}
				  if (data & 0x08)		// set read pointer
				    {
				      psg.adpcm_rptr = psg.adpcm_ptr;
				      psg.adpcm_firstread = 2;
				    }
				  if (data & 0x80)
				    {			// ADPCM reset
				    }
				  else
				    {			// Normal ADPCM utilisation
				    }
				  return;
				case 0xe:		// Set ADPCM playback rate
				  psg.adpcm_rate = 32 / (16 - (data & 15));
				  return;*/
           	}
            break;

        case 0x1C00: /* Expansion */
            break;
    }

#ifdef DEBUG
    error("write %02X to IO %04X (%08X)\n", data, address, h6280_get_reg(H6280_PC));
#endif
}
/*****************************/
 /*  yoyofr : from hugo       */
 /*****************************/   
int io_read_raw(int address)
{
    switch(address & 0x1C00)
    {
        case 0x0000: /* VDC */
            address&=3;
            if(address <= 0x0003) return (vdc_r(address));
            break;

        case 0x0400: /* VCE */
            address&=0x0407;
            if(address <= 0x0405) return (vce_r(address));
            break;

        case 0x0800: /* PSG */
            address&=0x080F;
            return psg_r(address);
            break;

        case 0x0C00: /* Timer */
            address&=0x0C01;
            //if(address == 0x0C00 || address == 0x0C01) return (H6280_timer_r(address & 1));
            //return (H6280_timer_r(0));
            return (H6280_timer_r(address & 1));
            break;

        case 0x1000: /* I/O */
            /*address&=0x1000;
            if(address == 0x1000)*/ return (input_r());
            break;

        case 0x1400: /* IRQ control */
            address&=0x1403;
            if(address == 0x1402 || address == 0x1403) return (H6280_irq_status_r(address & 1));
            break;

        case 0x1800: /* CD-ROM */
#ifdef DEBUG
            error("cdrom %04X\n", address);
#endif
			switch (address & 15)
			{
				case 3:
				{	
					static uint8 tmp_res = 0x02;
					tmp_res = 0x02 - tmp_res;
					save_bram = 0;
					return tmp_res | 0x20;
				}
/*				case 5:
					return 0x50;
			// TEST
					case 6:
					return 0x05;
				case 0x0A:
					if (!psg.adpcm_firstread)
					return PCM[psg.adpcm_rptr++];
				else
				{
					psg.adpcm_firstread--;
					return 0xFF;//NODATA;
				}
				case 0x0B:	// TEST
					return 0x00;
				case 0x0C:	// TEST
					return 0x01;	// 0x89
				case 0x0D:	// TEST
					return 0x00;*/
			}
            break;
        case 0x1C00: /* Expansion */
            break;
    }
#ifdef DEBUG
    error("read IO %04X (%08X)\n", address, h6280_get_reg(H6280_PC));
#endif    
    return 0xFF; //NODATA 
    //(0x00);
}

/**
 * Functions to access PCE hardware
 **/
//vdc reg
extern uint8 latch;

int return_value_mask_tab_0002[32] =
  {
    0xFF,
    0xFF,
    0xFF,
    0xFF, /* unused */
    0xFF, /* unused */
    0xFF,
    0xFF,
    0xFF,
    0xFF, /* 8 */
    0xFF,
    0x1F, /* A */
    0x7F,
    0x1F, /* C */
    0xFF,
    0xFF, /* E */
    0x1F,
    0xFF, /* 10 */
    0xFF,
    /* No data for remaining reg, assuming 0xFF */
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF
  };

int return_value_mask_tab_0003[32] =
  {
    0xFF,
    0xFF,
    0xFF,
    0xFF, /* unused */
    0xFF, /* unused */
    0x1F,
    0x03,
    0x03,
    0x01, /* 8 */	/* ?? */
    0x00,
    0x7F, /* A */
    0x7F,
    0xFF, /* C */
    0x01,
    0x00, /* E */
    0x00,
    0xFF, /* 10 */
    0xFF,
    /* No data for remaining reg, assuming 0xFF */
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF
  };


int return_value_mask_tab_0400[8] =
  {
    0xFF,
    0x00,
    0xFF,
    0x01,
    0xFF,
    0x01,
    0xFF, /* unused */
    0xFF  /* unused */
  };

int return_value_mask_tab_0800[16] =
  {
    0x03,
    0xFF,
    0xFF,
    0x0F,
    0xDF,
    0xFF,
    0x1F,
    0x9F,
    0xFF,
    0x83,
    /* No data for remainig reg, assuming 0xFF */
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF,
    0xFF
  };

int return_value_mask_tab_0c00[2] =
  {
    0x7F,
    0x01
  };

int return_value_mask_tab_1400[4] =
  {
    0xFF,
    0xFF,
    0x03,
    0x03
  };


//! Returns the useful value mask depending on port value
static int
return_value_mask(UInt16 A)
{
  if (A < 0x400) // VDC
    {
      if ((A & 0x3) == 0x02)
        {
          return return_value_mask_tab_0002[latch];
        }
      else
        if ((A & 0x3) == 0x03)
          {
            return return_value_mask_tab_0003[latch];
          }
        else
          return 0xFF;
    }

  if (A < 0x800) // VCE
    return return_value_mask_tab_0400[A & 0x07];

  if (A < 0xC00) /* PSG */
    return return_value_mask_tab_0800[A & 0x0F];

  if (A < 0x1000) /* Timer */
    return return_value_mask_tab_0c00[A & 0x01];

  if (A < 0x1400) /* Joystick / IO port */
    return 0xFF;

  if (A < 0x1800) /* Interruption acknowledgement */
    return return_value_mask_tab_1400[A & 0x03];

  /* We don't know for higher ports */
  return 0xFF;
}



int io_page_r(int address)
{     
	return io_read_raw(address);
 /* if ((address<0x800)||(address>=0x1800))  return io_read_raw(address);  
  
  int mask;
  uint8 temporary_return_value;   	
    
  mask = return_value_mask(address);
  temporary_return_value = io_read_raw(address);
  io_buffer = temporary_return_value | (io_buffer & ~mask);

  return io_buffer;*/
}

/*--------------------------------------------------------------------------*/
/* Input routines                                                           */
/*--------------------------------------------------------------------------*/

void input_w(uint8 data)
{
    joy_sel = (data & 1);
    joy_clr = (data >> 1) & 1;
}

uint8 input_r(void)
{
    uint8 temp = 0xFF;

    if(input.pad[joy_cnt] & INPUT_LEFT)   temp &= ~0x80;
    if(input.pad[joy_cnt] & INPUT_DOWN)   temp &= ~0x40;
    if(input.pad[joy_cnt] & INPUT_RIGHT)  temp &= ~0x20;
    if(input.pad[joy_cnt] & INPUT_UP)     temp &= ~0x10;
    if(input.pad[joy_cnt] & INPUT_RUN)    temp &= ~0x08;
    if(input.pad[joy_cnt] & INPUT_SELECT) temp &= ~0x04;
    if(input.pad[joy_cnt] & INPUT_B2)     temp &= ~0x02;
    if(input.pad[joy_cnt] & INPUT_B1)     temp &= ~0x01;

    if(joy_sel & 1) temp >>= 4;
    temp &= 0x0F;

    /* Set D6 for TurboGrafx-16, clear for PC-Engine */
    if(input.system & SYSTEM_TGX) temp |= 0x40;

    return (temp);
}

#ifdef FAST_MEM

void bank_reset(void)
{
    int i;
    for(i = 0; i < 8; i += 1)
    {
        read_ptr[i] = &rom[0x0000];
        write_ptr[i] = &dummy[0x0000];
    }
}

void bank_set(int bank, int value)
{
    /* ROM */
    if(value <= 0x7F) {
        read_ptr[bank] = &rom[(value << 13)];
        write_ptr[bank] = &dummy[0x0000];
        return;
    }

    /* CD RAM */
    if((value >= 0x80) && (value <= 0x87)) {
        read_ptr[bank] = write_ptr[bank] = &cdram[(value & 0x07) << 13];
        return;
    }

    /* RAM */
    if((value >= 0xF8) && (value <= 0xFB)) {
        read_ptr[bank] = write_ptr[bank] = &ram[(value & 0x03) << 13];
        return;
    }

    /* Backup RAM */
    if(value == 0xF7) {
        read_ptr[bank] = write_ptr[bank] = &bram[0x0000];
        return;
    }

    /* I/O page */
    if(value == 0xFF) {
        read_ptr[bank] = write_ptr[bank] = NULL;
        return;
    }

    /* Unknown page */
#ifdef DEBUG
    error("Map unknown page %02X to MMR #%d\n", value, bank);
#endif
    read_ptr[bank] = write_ptr[bank] = &dummy[0x0000];
}

#endif

